from hamlet.backend.engine import engine_store
from hamlet.backend.engine.common import (
    ENGINE_GLOBAL_NAME,
    ENGINE_DEFAULT_GLOBAL_ENGINE,
)
from hamlet.backend.engine.exceptions import (
    HamletEngineInvalidVersion,
    EngineStoreMissingEngineException,
)
from hamlet.env import HAMLET_GLOBAL_CONFIG


def setup_global_engine(engine_override):
    """
    Always make sure the global engine is installed and that we have at least one engine installed
    """

    try:
        global_engine = engine_store.get_engine(ENGINE_GLOBAL_NAME)
    except HamletEngineInvalidVersion:

        # If the global engine is old then we need to force it to be the latest
        engine_store.clean_engine(ENGINE_GLOBAL_NAME)

        global_engine = engine_store.get_engine(ENGINE_GLOBAL_NAME)
        global_engine.install()

    if not global_engine.installed or not global_engine.up_to_date():
        global_engine.install()

    if engine_store.global_engine is None:

        if engine_override:
            default_engine = engine_override
        else:
            default_engine = ENGINE_DEFAULT_GLOBAL_ENGINE

        try:
            engine_store.get_engine(default_engine)

        except EngineStoreMissingEngineException:
            engine_store.find_engine(default_engine).install()

        engine_store.global_engine = default_engine


def get_engine_env(engine_override):
    """
    Inject the hamlet engine environment variables into the context
    """

    if engine_override is not None:
        engine = engine_store.get_engine(engine_override)
    else:
        engine = engine_store.get_engine(ENGINE_GLOBAL_NAME)

    HAMLET_GLOBAL_CONFIG.engine_environment = engine.environment



import click
from click.types import StringParamType
import functools

from hamlet.command.common.config import Options


class CommaSplitParamType(StringParamType):
    envvar_list_splitter = ","

    def __repr__(self):
        return "STRING"


def common_cli_config_options(func):
    """Add common CLI config options to commands"""

    @click.option(
        "-p",
        "--profile",
        default=None,
        envvar="HAMLET_PROFILE",
        help="The name of the profile to use for configuration",
        show_envvar=True,
    )
    @click.pass_context
    @functools.wraps(func)
    def wrapper(ctx, *args, **kwargs):
        """
        Config file handling
        """
        opts = ctx.ensure_object(Options)
        profile = kwargs.pop("profile")
        opts.load_config_file(profile=profile)
        kwargs["opts"] = opts
        return ctx.invoke(func, *args, **kwargs)

    return wrapper


def common_logging_options(func):
    """Add commmon options for logging"""

    @click.option(
        "--log-level",
        envvar="GENERATION_LOG_LEVEL",
        type=click.Choice(
            ["fatal", "error", "warn", "info", "debug", "trace"], case_sensitive=False
        ),
        default="info",
        help="The minimum log event level",
        show_default=True,
        show_envvar=True,
    )
    @click.pass_context
    @functools.wraps(func)
    def wrapper(ctx, *args, **kwargs):
        """
        Logging Options for the command line
        """
        opts = ctx.ensure_object(Options)
        opts.log_level = kwargs.pop("log_level")
        kwargs["opts"] = opts
        return ctx.invoke(func, *args, **kwargs)

    return wrapper


def common_engine_options(func):
    """Add common options for the engine"""

    @click.option(
        "--engine",
        envvar="HAMLET_ENGINE",
        help="The name of the engine to use",
        show_envvar=True,
    )
    @click.pass_context
    @functools.wraps(func)
    def wrapper(ctx, *args, **kwargs):
        """
        Engine configuration options
        """
        opts = ctx.ensure_object(Options)
        opts.engine = kwargs.pop("engine")
        kwargs["opts"] = opts
        return ctx.invoke(func, *args, **kwargs)

    return wrapper


def common_generation_options(func):
    """Add commmon options for generation"""

    @click.option(
        "-p",
        "--generation-provider",
        envvar="GENERATION_PROVIDERS",
        help="plugins to load for output generation",
        default=["aws"],
        type=CommaSplitParamType(),
        multiple=True,
        show_default=True,
        show_envvar=True,
    )
    @click.option(
        "-f",
        "--generation-framework",
        help="output framework to use for output generation",
        default="cf",
        show_default=True,
    )
    @click.option(
        "-i",
        "--generation-input-source",
        help="source of input data to use when generating the output",
        default="composite",
        show_default=True,
    )
    @click.pass_context
    @functools.wraps(func)
    def wrapper(ctx, *args, **kwargs):
        """
        Logging Options for the command line
        """
        opts = ctx.ensure_object(Options)
        opts.generation_provider = kwargs.pop("generation_provider")
        opts.generation_framework = kwargs.pop("generation_framework")
        opts.generation_input_source = kwargs.pop("generation_input_source")
        kwargs["opts"] = opts
        return ctx.invoke(func, *args, **kwargs)

    return wrapper


def common_district_options(func):
    """Add Common options for district config"""

    @click.option(
        "--root-dir",
        envvar="ROOT_DIR",
        help="The root CMDB directory (default: CMDB current location)",
        show_envvar=True,
    )
    @click.option(
        "--tenant",
        envvar="TENANT",
        help="The tenant name to use (default: CMDB current location)",
        show_envvar=True,
    )
    @click.option(
        "--account",
        envvar="ACCOUNT",
        help="The account name to use",
        show_envvar=True,
    )
    @click.option(
        "--product",
        envvar="PRODUCT",
        help="The product name to use (default: CMDB current location)",
        show_envvar=True,
    )
    @click.option(
        "--environment",
        envvar="ENVIRONMENT",
        help="The environment name to use (default: CMDB current location)",
        show_envvar=True,
    )
    @click.option(
        "--segment",
        envvar="SEGMENT",
        help="The segment name to use (default: CMDB current location)",
        show_envvar=True,
    )
    @click.pass_context
    @functools.wraps(func)
    def wrapper(ctx, *args, **kwargs):
        """
        District options from cmd line or file
        """
        opts = ctx.ensure_object(Options)
        opts.root_dir = kwargs.pop("root_dir")
        opts.tenant = kwargs.pop("tenant")
        opts.account = kwargs.pop("account")
        opts.product = kwargs.pop("product")
        opts.environment = kwargs.pop("environment")
        opts.segment = kwargs.pop("segment")
        kwargs["opts"] = opts
        return ctx.invoke(func, *args, **kwargs)

    return wrapper



# NOTE: This file must remain valid python file in order to perform tests on it.
# NOTE: Imports can't be used inside a template block because all code will be merged into a single file.
# NOTE: The class is wrapped into function in order to make it testable. This adds an ability to provide parent classes
# at runtime, otherwise, the module will raise NameError


def CFNStructure(JSONValidator):
    class CFNStructure(JSONValidator):

        RESOURCES_KEY = "Resources"
        RESOURCE_TYPE_KEY = "Type"
        OUTPUT_KEY = "Outputs"

        def __resource(self, id, type):
            def validator():
                resources = self._body.get(self.RESOURCES_KEY, {})
                target = resources.get(id)
                assert target is not None, "cfn resource {} is missing".format(id)
                assert (
                    target[self.RESOURCE_TYPE_KEY] == type
                ), "cfn resource {}.{}!={}".format(id, self.RESOURCE_TYPE_KEY, type)

            validator.__name__ = "cfn resource {}.{} == {}".format(
                id, self.RESOURCE_TYPE_KEY, type
            )
            return validator

        def __output(self, id):
            def validator():
                output = self._body.get(self.OUTPUT_KEY, {})
                value = output.get(id)
                assert value is not None, "cfn output {} is missing".format(id)

            validator.__name__ = "cfn output {} exists".format(id)
            return validator

        def resource(self, id, type):
            self._validators.append(self.__resource(id, type))
            return self

        def output(self, id):
            self._validators.append(self.__output(id))
            return self

    return CFNStructure



from hamlet.backend.common import runner


def run(
    acceptance_tag=None,
    code_commits=None,
    list_full=None,
    segment_builds_dir=None,
    image_formats=None,
    list=None,
    registry_scopes=None,
    code_providers=None,
    code_repos=None,
    deployment_units=None,
    code_tags=None,
    update=None,
    verify=None,
    log_level=None,
    _is_cli=False,
    env={},
):
    opts = {
        "-a": acceptance_tag,
        "-c": code_commits,
        "-f": list_full,
        "-g": segment_builds_dir,
        "-i": image_formats,
        "-l": list,
        "-o": registry_scopes,
        "-p": code_providers,
        "-r": code_repos,
        "-s": deployment_units,
        "-t": code_tags,
        "-u": update,
        "-v": verify,
    }
    env = {"AUTOMATION_LOG_LEVEL": log_level, **env}
    runner.run(
        "manageBuildReferences.sh",
        [],
        opts,
        env,
        _is_cli,
        script_base_path_env="AUTOMATION_DIR",
    )
